#/bin/bash
# Test of reinstallation and downgrade support in PackageKit.
# 
# It runs a pkcon client as a testing user with several operations
# These operations are run in a loop with different policies applied.
# Policy environment is represented by array of flags (see $policies) where
# each flag denotes whether corresponding action in $policy_action_names is
# allowed for the user or not.
# 
# Influential environment variables:
#   INTERACTIVE         - Set to 1 to be asked for a permission to continue
#                         with next environment.
#   RESTART_PACKAGEKIT  - Set to 0 if the PackageKit should not be restarted.
#   TESTUSER            - Name of the user that shall run a client.
#   PASSWORD            - Password fot the ${TESTUSER}.
#
# Author: Michal Minar <miminar at redhat dot com>

. /usr/share/beakerlib/beakerlib.sh

INTERACTIVE="${INTERACTIVE:-0}"
RESTART_PACKAGEKIT=${RESTART_PACKAGEKIT:-1}
TESTUSER="${TESTUSER:-testuser}"
PASSWORD="${PASSWORD:-mypass}"
POLKIT_RULE_FILE="00-packagekit-test.rules"
DIST=`rpm -q --qf '%{RELEASE}\n' kernel | tail -n 1 | \
    sed -n 's/.*\.\([^.]\+\)$/\1/p'`
PKCON_RUNNER=./runpkcon.exp
PC="/tmp/runpkcon.exp $TESTUSER $PASSWORD"

TEST=PackageKit

policy_action_names=( \
    "package-install" \
    "package-install-untrusted" \
    "package-reinstall" \
    "package-downgrade" \
)

set -a policy_action_full_names
for i in ${policy_action_names[*]}; do
    policy_action_full_names+=("org.freedesktop.packagekit.$i")
done

EXIT_CODE_ALREADY_INSTALLED=7
EXIT_CODE_AUTHENTICATION_FAILED=7

REPOSTABLE="pkgkit-test-repo-stable"
REPOUPDATES="pkgkit-test-repo-updates"

PKG="pkgkit-test-pkg"
PKGVERSION="1.2.3"
PKGNEVRA="$PKG-$PKGVERSION-2.${DIST}.noarch"
PKGNEVRA_OLD="$PKG-$PKGVERSION-1.${DIST}.noarch"
function make_pkg_id() {
    echo "$1" | sed "s/\(.\+\)-\(\(\d\+:\)\?[^-]\+-[^-]\+\)\.\([^.]\+\)/\1;\2;\4;$2/"
}
PKGID=`make_pkg_id $PKGNEVRA $REPOUPDATES`
PKGID_OLD=`make_pkg_id $PKGNEVRA_OLD $REPOSTABLE`

function policy_flags2str() {
    # Get a list of policy actions from glags.
    # Arguments:
    #   POLICY_FLAGS - Binary string of the same length as policy_action_names
    #                  array. Flag's position in the string matches index of
    #                  correcponding action in policy_action_names array.
    # Returns: list of allowed actions separated with spaces
    flags="$1"
    first_printed=false
    for ((i=0; i<${#policy_action_names[@]}; i++)); do
        if [ "${flags:i:1}" = 1 ]; then
            $first_printed && printf " "
            printf "${policy_action_names[$i]}"
            first_printed=true
        fi
    done
    printf "\n"
}

function write_polkit_rules() {
    # Write a policy kit rule file named ${POLKIT_RULE_FILE}.
    # If flag is set, user needs to authenticate himself for corresponding
    # action. Otherwise he's turned down.
    # Arguments:
    #   POLICY_ENV_FLAGS
    flags="$1"
    if [ ${#policy_action_names[@]} != ${#flags} ]; then
        echo "Unexpected length of policy env flags!" >&2
        exit 1
    fi
    ( \
        printf "polkit.addRule(function(action, subject) {\n"; \
        for ((i=0; i<${#policy_action_names[@]}; i++)); do
            flag=${flags:$i:1}
            printf "    if (action.id == \"${policy_action_full_names[$i]}\" &&\n"
            printf "            subject.user == \"$TESTUSER\")\n"
            printf "    {\n"
            printf "        return polkit.Result."
            if [ "$flag" = 1 ]; then
                printf "AUTH_SELF;\n"
            else
                printf "NO;\n"
            fi
            printf "    }\n"
            shift
        done
        printf "});\n"
    ) >/etc/polkit-1/rules.d/$POLKIT_RULE_FILE
}

function check_authentications() {
    # Parse to stdout log file of pkcon and make sure it contains expected
    # authorization entries as denoted by given flags.
    # Arguments:
    #  POLICY_ENV_FLAGS     - Policy kit environment flags.
    #  PKCON_FLAGS          - Flags denote actions requested by a client.
    #  LOG_FILE             - stdout log file of executed client.
    env_flags="$1"
    pkcon_flags="$2"
    log_file="$3"
    exp_number_of_auths=0
    if [ ${#policy_action_names[@]} != ${#env_flags} ]; then
        echo "Unexpected length of policy env flags!" >&2
        exit 1
    fi
    if [ ${#policy_action_names[@]} != ${#pkcon_flags} ]; then
        echo "Unexpected length of pkcon flags!" >&2
        exit 1
    fi
    for ((i=0; i<${#env_flags}; i++)); do
        action="${policy_action_names[$i]}"
        if [ ${pkcon_flags:$i:1}${env_flags:$i:1} = 11 ]; then
            let "exp_number_of_auths++"
            rlAssertGrep "${policy_action_full_names[$i]}\s\+" $log_file
        elif [ ${pkcon_flags:$i:1} = 0 ]; then
            rlAssertNotGrep "${policy_action_full_names[$i]}\s\+" $log_file
        elif [ ${pkcon_flags:$i:1}${env_flags:$i:1} = 10 ]; then
            break
        fi
    done
    rlAssertEquals "Checking number of authentications" \
        `grep -c AUTHENTICATING $log_file` $exp_number_of_auths
}

function make_testing_repos() {
    # Create testing repositories.
    # Returns: Absolute path to a directory with created repositories.
    set -e
    local rpmbuilddir repodir rpmname rpmnames
    rpmbuilddir=`mktemp -d`
    repodir=`mktemp -d`
    declare -a rpmnames
    for dir in SOURCES SPECS BUILD RPMS; do
        mkdir "$rpmbuilddir/$dir"
    done

    pushd $rpmbuilddir >/dev/null
    for release in 1 2; do
        pushd SOURCES >/dev/null
            mkdir $PKG-$PKGVERSION
            echo "This is a test package $PKG v$PKGVERSION" >$PKG-$PKGVERSION/README
            tar -czf $PKG-${PKGVERSION}{.tar.gz,}
            rm -rf $PKG-$PKGVERSION
        popd >/dev/null    # SOURCES

        cat >SPECS/${PKG}.spec <<-EOF
	Name:     $PKG
	Epoch:    0
	Version:  $PKGVERSION
	Release:  $release%{?dist}
	License:  LGPLv2+
	Source0:  %{name}-%{version}.tar.gz
	BuildArch: noarch
	Summary: This is a test package $PKG
	
	%description
	Test package for PackageKit.
	
	%prep
	%setup -q
	
	%build
	# no build necessary
	
	%install
	    mkdir -p \$RPM_BUILD_ROOT%{_docdir}/%{name}
	    cp README \$RPM_BUILD_ROOT%{_docdir}/%{name}/
	
	%files
	%doc %{_docdir}/%{name}/README
	
	%changelog
	* Mon Nov 04 2013 Michal Minar <miminar@redhat.com> $PKGVERSION-$release
	- initial commit
	EOF
        rpmbuild --quiet -ba -D "_topdir $rpmbuilddir" SPECS/${PKG}.spec
        rpmname=`rpm -q --specfile SPECS/${PKG}.spec`
        rpmnames+=("${rpmname}.rpm")
    done
    mkdir -p $repodir/$REPOSTABLE
    mkdir -p $repodir/$REPOUPDATES
    cp RPMS/noarch/${rpmnames[0]} $repodir/$REPOSTABLE
    cp RPMS/noarch/${rpmnames[1]} $repodir/$REPOUPDATES
    popd >/dev/null   # rpmbuilddir

    rm -rf $rpmbuilddir
    for name in $REPOSTABLE $REPOUPDATES; do
        cat >/etc/yum.repos.d/${name}.repo <<-EOF
	[$name]
	name=$name
	baseurl=file://$repodir/$name
	enabled=0
	gpgcheck=0
	skip_if_unavailable=1
	EOF
    done
    set +e
    createrepo --quiet $repodir/$REPOSTABLE
    createrepo --quiet $repodir/$REPOUPDATES
    echo $repodir
}

if [ "$UID" != 0 ]; then
    echo "This test script must be run under root." >&2
    exit 1
fi

if ! id $TESTUSER >/dev/null 2>&1; then
    echo "Please create test user \"$TESTUSER\" account." >&2
    exit 1
fi

# Array of flags indicating what action are allowed for test user.
# Flags represent following actions:
#       org.freedesktop.packagekit.package-install
#       org.freedesktop.packagekit.package-install-untrusted
#       org.freedesktop.packagekit.package-reinstall
#       org.freedesktop.packagekit.package-downgrade
policies=( "1111" "0111" "0110" "0101" "0100" "0011" "0010" "0001" "0000" )

rlJournalStart

rlPhaseStartSetup

    rlLogInfo "Creating local testing repositories."
    repodir=`make_testing_repos`
    if ! [ -d "$repodir" ]; then
        echo "Failed to create testing repositories!" >&2
        exit 1
    fi
    rlLogInfo "Repositories created in $repodir"
    for repo in $REPOSTABLE $REPOUPDATES; do
        rlLogInfo "  $repo with:"
        find "$repodir/$repo" -type f -name '*.rpm' | while read name;
        do
            rlLogInfo "    `basename $name`"
        done
    done

    rlLogInfo "Checking repositories."
    if [ `yum -q repolist 2>/dev/null | \
            grep '^pkgkit-test-repo-\(stable\|updates\)' | wc -l` -lt 2 ];
    then
        rlLogInfo "Disabling other repositories."
        yum -q repolist 2>/dev/null | tail -n +1 | \
            grep -v '^pkgkit-test-repo-\(stable\|updates\)' | \
            sed -n -e '/^repolist/d' -e 's,^\([^ /]\+\).*,\1,p' | \
            xargs -r yum-config-manager -q --disable >/dev/null
        rlLogInfo "Enabling testing repositories."
        yum-config-manager -q --enable \
            $REPOSTABLE $REPOUPDATES >/dev/null
        rlLogInfo "Restarting PackageKit."
        if [ $RESTART_PACKAGEKIT = 1 ]; then
            systemctl kill packagekit
            systemctl start packagekit
        fi
    fi

    cp $PKCON_RUNNER /tmp/

rlPhaseEnd

for policy_flags in ${policies[*]}; do
    allow_install="${policy_flags:0:1}"
    allow_install_untrusted="${policy_flags:1:1}"
    allow_reinstall="${policy_flags:2:1}"
    allow_downgrade="${policy_flags:3:1}"

    write_polkit_rules $policy_flags
    rlPhaseStartTest "Testing pkcon with allowed policy actions: $(policy_flags2str $policy_flags)"

    rlPhaseStartTest "Test normal installation with flags: $policy_flags"
        rpm -e $PKG 2>/dev/null
        pkcon refresh -p
        rlRun -s "$PC install $PKG </dev/null" \
            `[ "$allow_install_untrusted" = 1 ] && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED`
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if [ "$allow_install_untrusted" = 1 ]; then
            rlAssertGrep "^Status:\s*Finished" $rlRun_LOG
            rlAssertGrep "^Installed.*$PKGNEVRA" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "rpm -qa '$PKG'" 0 "Package is installed."
        else
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0100 $rlRun_LOG
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test install newer package with flags: $policy_flags"
        rpm -e $PKG 2>/dev/null
        rlRun "yum install -q -y $PKGNEVRA_OLD" 0 "Install older package."
        pkcon refresh -p
        rlRun -s "$PC install $PKG </dev/null" \
            `[ "$allow_install_untrusted" = 1 ] && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED` \
            "Update the package."
        rlAssertGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if [ "$allow_install_untrusted" = 1 ]; then
            rlAssertGrep "^Updated.*$PKG_NEVRA" $rlRun_LOG
            rlAssertGrep "^Cleaned up*$PKG_NEVRA_OLD" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "rpm -qa $PKGNEVRA" 0 "Package is installed."
        else
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0100 $rlRun_LOG
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test disallowed reinstallation with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKGNEVRA"
        fi
        rlRun -s "$PC install \"$PKGID\"" \
            $EXIT_CODE_ALREADY_INSTALLED
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        rlAssertNotGrep "^\(Updated\|Installed\)" $rlRun_LOG
        rlAssertNotGrep "error.*higher version" $rlRun_LOG
        rlAssertGrep "error.*is alr\?eady installed" $rlRun_LOG
        rlAssertEquals "Checking number of authentications" \
            `grep -c AUTHENTICATING $rlRun_LOG` 0
        rlRun "rpm -q $PKG" 0 "Package is still installed."
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test disallowed reinstallation with downgrades allowed with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKGNEVRA"
        fi
        pkcon refresh -p
        rlRun -s "$PC --allow-downgrade install \"$PKGID\"" \
            $EXIT_CODE_ALREADY_INSTALLED
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        rlAssertNotGrep "^\(Updated\|Installed\)" $rlRun_LOG
        rlAssertNotGrep "error.*higher version" $rlRun_LOG
        rlAssertGrep "error.*is alr\?eady installed" $rlRun_LOG
        rlAssertEquals "Checking number of authentications" \
            `grep -c AUTHENTICATING $rlRun_LOG` 0
        rlRun "rpm -q $PKG" 0 "Package is still installed."
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test reinstallation with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKGNEVRA" 0
        fi
        rlRun "rpm -ql $PKG | grep README | xargs rm" 0 "Modifying installed file."
        shall_succeed=`[ "$allow_install_untrusted$allow_reinstall" = 11 ] \
            && echo true || echo false`
        pkcon refresh -p
        rlRun -s "$PC --allow-reinstall install \"$PKGID\" </dev/null" \
            `$shall_succeed && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED` \
            "Reinstall modified package."
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if $shall_succeed; then
            rlAssertGrep "^Reinstalled.*$PKG_NEVRA" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "rpm -V $PKG"
            rlRun '[ -e `rpm -ql $PKG | grep README` ]' \
                0 "Removed file is restored."
        else 
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0110 $rlRun_LOG
        [ $? = 0 ] || rpm -e $PKG 2>/dev/null
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test disallowed downgrade with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKG" 0
        fi
        pkcon refresh -p
        rlRun -s "$PC install \"$PKGID_OLD\"" $EXIT_CODE_ALREADY_INSTALLED
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        rlAssertNotGrep "^\(Updated\|Installed\)" $rlRun_LOG
        rlAssertGrep "error.*higher version .* of package .* is already installed" \
            $rlRun_LOG
        rlAssertEquals "Checking number of authentications" \
            `grep -c AUTHENTICATING $rlRun_LOG` 0
        rlRun "[ \`rpm -qa $PKG\` == $PKGNEVRA ]" 0 \
            "Installed package stays the same."
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test disallowed downgrade with reinstallation allowed with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKG" 0
        fi
        pkcon refresh -p
        rlRun -s "$PC --allow-reinstall install \"$PKGID_OLD\"" \
            $EXIT_CODE_ALREADY_INSTALLED
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        rlAssertNotGrep "^\(Updated\|Installed\)" $rlRun_LOG
        rlAssertGrep "error.*higher version .* of package .* is already installed" \
            $rlRun_LOG
        rlAssertEquals "Checking number of authentications" \
            `grep -c AUTHENTICATING $rlRun_LOG` 0
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test downgrade with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKG" 0
        fi
        shall_succeed=`[ "$allow_install_untrusted$allow_downgrade" = 11 ] \
            && echo true || echo false`
        pkcon refresh -p
        rlRun -s "$PC --allow-downgrade install \"$PKGID_OLD\" </dev/null" \
            `$shall_succeed && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED` \
            "Downgrade the package."
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if $shall_succeed; then
            rlAssertGrep "^Updated" $rlRun_LOG
            rlAssertGrep "^Cleaned up.*$PKGNEVRA" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "[ \`rpm -qa $PKG\` == $PKGNEVRA_OLD ]" 0 \
                "Older package is installed."
        else
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0101 $rlRun_LOG
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test reinstall with both flags set with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKG" 0
        fi
        rlRun "[ \`rpm -qa $PKG\` == $PKGNEVRA ]" 0
        shall_succeed=`echo $policy_flags | grep -q '^[0-1]111$' && echo true || echo false`
        pkcon refresh -p
        rlRun -s "$PC --allow-reinstall --allow-downgrade install \"$PKGID\" </dev/null" \
            `$shall_succeed && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED` \
            "Reinstall the package."
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if $shall_succeed; then
            rlAssertGrep "^Reinstalled" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "[ \`rpm -qa $PKG\` == $PKGNEVRA ]" 0 \
                "Package is reinstalled."
        else
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0111 $rlRun_LOG
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseStartTest "Test downgrade with both flags set with flags: $policy_flags"
        if ! rpm --quiet -q $PKGNEVRA ; then
            rpm -e $PKG 2>/dev/null
            rlRun "yum install -q -y $PKG" 0
        fi
        shall_succeed=`echo $policy_flags | grep -q '^[0-1]111$' && echo true || echo false`
        pkcon refresh -p
        rlRun -s "$PC --allow-reinstall --allow-downgrade install \"$PKGID_OLD\" </dev/null" \
            `$shall_succeed && echo 0 || echo $EXIT_CODE_AUTHENTICATION_FAILED` \
            "Downgrade the package to $PKGNEVRA_OLD"
        rlAssertNotGrep "^Transaction:\s*Resolving" $rlRun_LOG
        rlAssertGrep "^Transaction:\s*Installing" $rlRun_LOG
        if $shall_succeed; then
            rlAssertGrep "^Updated" $rlRun_LOG
            rlAssertGrep "^Cleaned up.*$PKGNEVRA" $rlRun_LOG
            rlAssertNotGrep "error" $rlRun_LOG
            rlRun "[ \`rpm -qa $PKG\` == $PKGNEVRA_OLD ]" 0 \
                "Older package is installed."
        else
            rlAssertGrep "Fatal error: Failed to obtain authentication." $rlRun_LOG
        fi
        check_authentications $policy_flags 0111 $rlRun_LOG
        rm $rlRun_LOG
    rlPhaseEnd

    rlPhaseEnd

    if [[ "$INTERACTIVE" =~ ^[yY1] ]]; then
        printf "Continue? [y/n] "
        read cont
        if [[ "$cont" =~ ^(0|[nN]) ]]; then
            break
        fi
    fi

done

rlPhaseStartCleanup
    rlLogInfo "Removing testing packages."
    yum -q remove -y "$PKG" 2>/dev/null
    rm /etc/polkit-1/rules.d/$POLKIT_RULE_FILE
    rm /tmp/$PKCON_RUNNER
    rlLogInfo "Deleting testing repositories."
    rm -rf $repodir
    rm /etc/yum.repos.d/{$REPOSTABLE,$REPOUPDATES}.repo
rlPhaseEnd

rlJournalEnd

# ex: et ts=4 sw=4
